/** Defines.java

	This program converts a Java properties file into a 
	.h file with #defines file suitable for the C preprocessor.
	It also produces a .sed file with s/.../.../g commands
	for the same expansions.

	Lines beginning with # are deleted.
	Other lines are in the form

		property.name = value string extending to end-of-line

	These lines are converted for the .h file as
		#define PROPERTY_NAME value string extending to end-of-line
	Note that the property name is converted to all caps and 
	periods are converted to underline.
	The conversion for the .sed file is 
		s/PROPERTY_NAME/value string extending to end-of-line/g

	The first argument is the path to the properties files,
	ending in ".properties"  The output files are the same but 
	have the extension .h and .sed.

	After reading in the .properties file with java.util.Properties.load(),
	the property values are scanned for subsequences ${...}.  If the 
	contents name a property, the value of that property is substituted
	for the entire subsequence.  Prior to the substitution, the value
	is itself scanned for possible substitutions by this rule.  
	Recursion is detected by limiting this process to ten levels.
	The sequence $${ is converted to ${
*/

import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Properties;


public class Defines {

	static java.util.Properties props = new java.util.Properties();


// write a single line to a BufferedWriter
private static void writeLine(java.io.BufferedWriter f, String s) 		throws java.io.IOException {
	f.write(s, 0, s.length());
	f.newLine();
}


// choose a delimiter which is not among the characters of a
private static char chooseDelim(String a) {
	String candidates = 
		"'\":;.,/^%$#@!~`><][}{|+=-)(*&?"
		+ "1234567890qwertyuiopasdfghjklzxcvbnm"
		+ "QWERTYUIOPLKJHGFDSAZXCVBNM_";

	while (candidates.length() > 0) {
		char thisChar = candidates.charAt(0);
		if (a.indexOf(thisChar) < 0)
			return thisChar;
		candidates = candidates.substring(1);
	}

	throw new IllegalArgumentException(
		"no character available as delimiter");

}


// in val, expand ${...} to value of ...
// cnt increments for expansion depth and does not exceed 10
//	(cnt does NOT increment when expanding a tail of a string)
// the algorithm is highly inefficient and recursive, 
//	not too many ${...} are expected
private static String expand(String val, int cnt) {
	if (cnt >= 10) return val;	// avoid recursion
	int dolloc = val.indexOf("${");	// location of first "${"
	if (dolloc < 0) return val;	// no "${"
	if (dolloc > 0 && val.charAt(dolloc-1) == '$')
		// "$${" -> "${"
		return val.substring(0,dolloc-1) + "$"
			+ expand(val.substring(dolloc+1), cnt);
	int rightloc = val.indexOf('}', dolloc);
	if (rightloc < 0) return val;	// no trailing '}'
	
	// expand ${...}
	String name = val.substring(dolloc+2, rightloc);
	return val.substring(0, dolloc)
		+ expand((String)props.get(name), cnt+1)
		+ expand(val.substring(rightloc+1), cnt);
}

public static void main(String args[]) throws java.io.IOException {

	// setup filenames and file objects
	//
	String inFileName = args[0];
	String hFileName, sedFileName;
	if ( ! inFileName.endsWith(".properties")) 
		throw new IllegalArgumentException(
				"file name must ends with \".properties\"");
	hFileName = inFileName.substring(0, 
			inFileName.length()-".properties".length()) + ".h";
	sedFileName = inFileName.substring(0, 
			inFileName.length()-".properties".length()) + ".sed";
	System.out.println(inFileName + " => " + hFileName 
			+ " & " + sedFileName);

	java.io.FileInputStream inFile 
			= new java.io.FileInputStream(inFileName);
	java.io.BufferedWriter hFile 
			= new java.io.BufferedWriter(
				new java.io.FileWriter(hFileName));
	java.io.BufferedWriter sedFile 
			= new java.io.BufferedWriter(
				new java.io.FileWriter(sedFileName));

	props.load(inFile);
	inFile.close();
	java.util.Enumeration pe = props.keys();

	while (pe.hasMoreElements()) {
		String key = (String)pe.nextElement();
		String keyOut = key.replace('.', '_').toUpperCase();
		String value = expand((String)props.get(key), 0);
		char delim = chooseDelim(keyOut+value);

		writeLine(hFile, "#define " + keyOut + " " + value);
		writeLine(sedFile, "s" + delim + keyOut 
					+ delim + value + delim + "g");
	}

	hFile.close();
	sedFile.close();
}

}

