
package edu.cmu.sei.gee;

import java.util.Properties;
import java.io.FileInputStream;
import java.io.IOException;

/*	This is file GeeProperties.j.  
	It is preprocessed to produce GeeProperties.java.

====>	DO NOT EDIT GeeProperties.java.

*/

/**
	Provides access to the gee.properties file and alternatives.
<p>
	A GeeProperties object provides a connection to the gee.properties
	file, which is fetched once per process and converted into a 
	java.util.Properties object.  Properties can be fetched with the
	get or getPassword methods.
<p>
	Properties are fetched from GEEDEST/config/gee.properties,
	where GEEDEST is by default the place where GEE was originally
	installed.  If it has been moved elsewhere, there must be
	a java system property gee.destination.directory giving the 
	location of the GEE tree and this value is used for GEEDEST.
	CHECKING THE SYSTEM PROPERTY CAN BE A SECURITY HOLE.
<p>
	For debugging, the system property gee.use.sandbox is
	checked.  If set, its value is used as the full pathname
	of the root of the source tree.  In particular, gee.properties
	is in the config directory of that tree.  When gee.use.sandbox
	is set, programs should have no need to know where the 
	destination directory is.
	THIS IS A SECURITY HOLE.

*/
public class GeeProperties {
	private static String geeDirectory;
	private static Properties gProps = new Properties();
	static {  
		String sandbox = System.getProperty("gee.use.sandbox");
		String geePropsFile;
		if (sandbox != null) {
			geeDirectory = "/dev/null";
			geePropsFile = sandbox + "/config/gee.properties";
		}
		else {
			// the word in all caps is pre-processing into
			// the path name for the root of Gee installation.
			geeDirectory 
			    = System.getProperty("gee.destination.directory", 
				"GEE_DESTINATION_DIRECTORY");
			geePropsFile 
			    = geeDirectory + "/config/gee.properties";
		}

		// read gee.properties into tProps
		Properties tProps = new Properties();
		try {
			FileInputStream fin = new FileInputStream(geePropsFile);
			tProps.load(fin);
			fin.close();
		} catch(IOException ex) {
			// default to using System.properties (which is useless)
			tProps = System.getProperties();
		}

		// copy tProps to gProps, expanding ${...}
		java.util.Enumeration pe = tProps.keys();
		while (pe.hasMoreElements()) {
			String key = (String)pe.nextElement();
			String value = (String)tProps.get(key);
			value = expand(value, tProps, 0);
			gProps.setProperty(key, value);
		}
	}


/** Returns the location of where the Gee tree was installed.
	The orignal value is where the tree is installed.
	If the tree is moved, the system property
	gee.destination.directory can be set to say where
	gee is; this value overrides the original value.
*/
public static String geeDir() { return geeDirectory; }

/** Returns a password property.
	The default implementation just fetches it like any other
	property.  A sub-class may establish some more secure mechanisms.
*/
public String getPassword(String prop) { return gProps.getProperty(prop); }

/** Gets the value of a property. 
	@returns null if the property is undefined.
*/
public String get(String prop) { return gProps.getProperty(prop); }

/** Gets the value of a property, or returns a default value.
	@returns default if the property is undefined.
*/
public String get(String prop, String defaultValue) 
	{ return gProps.getProperty(prop, defaultValue); }

/** Gets the value of a property.
	Unlike get(), no GeeProperties object needs to be created.
*/
public static String valueFor(String prop) 
	 { return gProps.getProperty(prop); }

/** Gets the value of a property, or returns a default value.
	Unlike get(), no GeeProperties object needs to be created.
	@returns default if the property is undefined.
*/
public static String valueFor(String prop, String defaultValue) 
	{ return gProps.getProperty(prop, defaultValue); }

/** Constructs a path name depending on whether a test is in progress.
<p>
	If the property gee.use.sandbox is set <em>in the system 
	properties</em>, this function returns the concatenation
	of the value of this property and the second argument. 
	Otherwise it returns the concatenation of the root of the 
	installation tree with the first argument.
<p>
	When testing, an engineer may wish to utilize files located 
	in a local sandbox source tree rather than those in the 
	installed destination tree.  The method makes that possible.
	The trick is that when testing the engineer uses -D to set the property 
<pre>
		gee.use.sandbox
</pre>
	to have as its value the root of the sandbox source tree.
	
@param destLoc Location postfix to be appended to the 
			root of the gee destination directory.
@param testLoc Location postfix to be appended to the
			root of the current build tree.
*/
public String getCurrentLocation(String destLoc, String testLoc) {
	String sandbox = System.getProperty("gee.use.sandbox");
	if (sandbox != null) 
		return sandbox + "/" + testLoc;
	else return GeeProperties.geeDir() + "/" + destLoc;
}

/**
	Returns an enumeration of all the defined keys.
*/
public java.util.Enumeration keys()
	{ return gProps.keys(); }


// in val, expand ${...} to value of ...
// cnt increments for expansion depth and does not exceed 10
//	(cnt does NOT increment when expanding a tail of a string)
// the algorithm is highly inefficient and recursive, 
//	not too many ${...} are expected
private static String expand(String val, Properties props, int cnt) {
	if (cnt >= 10) return val;	// avoid recursion
	int dolloc = val.indexOf("${");	// location of first "${"
	if (dolloc < 0) return val;	// no "${"
	if (dolloc > 0 && val.charAt(dolloc-1) == '$')
		// "$${" -> "${"
		return val.substring(0,dolloc-1) + "$"
			+ expand(val.substring(dolloc+1), props, cnt);
	int rightloc = val.indexOf('}', dolloc);
	if (rightloc < 0) return val;	// no trailing '}'
	
	// expand ${...}
	String name = val.substring(dolloc+2, rightloc);
	return val.substring(0, dolloc)
		+ expand((String)props.get(name), props, cnt+1)
		+ expand(val.substring(rightloc+1), props, cnt);
}


} // end class GeeProperties
