package edu.cmu.sei.gee;

import edu.cmu.sei.gee.GeeProperties;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.util.Date;
import java.util.SimpleTimeZone;
import java.text.SimpleDateFormat;

/**
	This class provides a simple interface for logging messages.  
	By default it connects to a logging server.  If the
	server is down, messages are logged to a local file whose
	location is sent in a mail message to an administrator.
<p>
	The default log is available for simple logging.  For more
	complex cases, separate log objects can be created.
<p>
	Each entry is time stamped.  To save space, the full
	date and time are written in an initial entry and subsequent
	entries are stamped with the number of seconds since that 
	initial entry.  All times are GMT (Greenwich).
*/
public class Log {

	private static String server
		= GeeProperties.valueFor("gee.log.server");
	// xxx -- need to define gee.log.server in gee.properties

	private static String mailBackup 
		= GeeProperties.valueFor("gee.log.mailbackup");
	// xxx -- need to define gee.log.mailbackup in gee.properties

	private static SimpleDateFormat digitsFmt
		= new SimpleDateFormat("yyyyMMddkkmm");
	private static SimpleDateFormat logFmt
		= new SimpleDateFormat("yyyy MMM dd kk:mm");
	static {
		// do times in GMT
		String tzIDs[] = SimpleTimeZone.getAvailableIDs(0);
		SimpleTimeZone tz = new SimpleTimeZone(0, tzIDs[0]);
		logFmt.setTimeZone(tz);
		digitsFmt.setTimeZone(tz);
	}

	/** a place to do logging by default.
		The log will be on the server and 
		will be named "StdyyyMMddkkmm" where
		everything after "Std" is the digits of the
		current date/time (GMT).
	*/
	public static Log std
		= new Log("Std" + digitsFmt.format(new Date()));

	// object fields

	private String thisServer = null;
	// xxx -- add private fields holding connection to server 
	private String logFileName = null;
	private PrintStream logFile = null;

	private long startTime;	// when the log was opened

/** Creates a log stream named for the current time.  
	The name will be in the format yyyyMMddkkmm
	(year, month, day, hour(24), minute)
*/
public Log() { this(server, digitsFmt.format(new Date())); }

/** Creates a log stream with a given name.
*/
public Log(String logName) { this(server, logName); }

/** Create a log stream with a given name and on a given server.
*/
public Log(String serverAddr, String logName) {
	thisServer = serverAddr;
	logFileName = logName;

	if (thisServer != null) {
		// xxx -- connect to server
		// xxx -- if the connection fails,
		// xxx --	send mail to mailBackup 
		// xxx --	and set thisServer = null
		thisServer = null;
	}
	if (thisServer == null)
		useLogFile();
	Date now = new Date();
	startTime = now.getTime();
	enter("Start logging -- " + logFmt.format(now));
}

/** Sets the mailing address for a message if the logging service fails.
	(One message is sent.  Log entries are written to a local file.)
*/
public static void setMail(String mailTo) 
	{ mailBackup = mailTo; }

private void useLogFile() {
	// xxx -- break logFileName into path and fileName
	// xxx -- (cannot use java.io.File's separator, it's system dep't)
	// xxx -- if path is empty, 
	// xxx -- use	java.io.File.createTempFile(fileName+"#.log")
	// xxx -- else  java.io.File.createTempFile(fileName+"#.log", path)
	try {
		logFile = new PrintStream(
			new FileOutputStream(logFileName + ".log"));
	}
	catch (IOException ex) {
		// xxx -- send mail 
		logFile = null;
	}
}


/** Make an entry in the log.
	The entry will be preceded by a timestamp consisting
	of the number of seconds since the Log was started.
	(The first entry, with a small or zero timestamp, 
	will list the actual date and time, GMT.)
*/
public synchronized void enter(String msg) {
	msg = "" + ((new Date().getTime()-startTime)/1000) + "  " + msg;
	if (thisServer != null) {
		// xxx -- send msg to server
		// xxx -- if the send fails,
		// xxx --	send mail to mailBackup 
		// xxx --	and set to use a log file
		useLogFile();	// (sets thisServer = null)
	}
	if (thisServer == null && logFile != null) {
		logFile.print(msg);
		logFile.write('\n');	// keep log on Unix box
		if (logFile.checkError()) {
			logFile = null;
			// xxx -- send mail
		}
	}
}

public boolean loggingOK()
	{ return thisServer != null || logFile != null; }

} // end class Log




