/* 
 * rcsmakelinks
 *
 * Makes symbolic links in an object directory to entries in a source file directory.
 *
 * Deletes any existing links that do not correspond to links in the source directory.
 *
 * Deletes any files that should be links and replaces them with links.
 *
 * Does not makes links to entries which are directories, begin with '.', or end
 * with ".private"; deletes existing entries in the object directory with these
 * undesirable charactersitics.
 *
 * Usage: rcsmakelinks [-R] srcdir objdir
 * Exit: 0 if successful; 1 if anything went wrong.
 */

#ifdef _IBMR2
#define _ALL_SOURCE
#endif
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <stdio.h>

#if (!defined(_IBMR2) && !defined(hpux) && !defined(mips) && !defined(__svr4__))
#include <sys/dir.h>
#define dirent direct
#else
#include <dirent.h>
#endif
#ifndef _S_IFDIR
#define _S_IFDIR S_IFDIR
#endif

#include <errno.h>
extern int errno;
#include <ctype.h>

extern void free();
extern char *malloc();
extern int chdir();
extern char *rindex();
extern int readlink();
extern int unlink();
extern int symlink();
extern int getwd();

#define HASHSIZE 2000	/* hash table size for source directory entries */
static struct h_ent {char *Name; int Forgotten;}
	HashTable[HASHSIZE] = {{NULL, 0}, {NULL, 0}};

static void Stop(f, a1, a2, a3, a4, a5, a6)
     char *f;
{
    char format[1000];
    sprintf(format, "rcsmakelinks: %s; aborted\n", f);
    fprintf(stderr, format, a1, a2, a3, a4, a5, a6);
    exit(1);
}

static void BTW(f, a1, a2, a3, a4, a5, a6)
     char *f;
{
    char format[1000];
    sprintf(format, "rcsmakelinks: %s\n", f);
    fprintf(stderr, format, a1, a2, a3, a4, a5, a6);
}


static void ClearHashTable()
{
    register int i;
    
    for (i = 0; i < HASHSIZE; i++)  {
	if (HashTable[i].Name)  {
	    free(HashTable[i].Name);
	    HashTable[i].Name = 0;
	}
	HashTable[i].Forgotten = 0;
    }
}

static int HashString(s, size)
    register char *s;
    unsigned int size;
{
    register unsigned int sum;
    register int n;
    for (sum = 0, n = strlen(s), s += n-1; n--; s--)
	sum = (sum*31) + (*s-31);
    return (sum%size);
}

static char *NewString(s)
char *s;
{
    char *p = (char *)malloc(strlen(s)+1);
    strcpy(p, s);
    return(p);
}

static void Remember(name)
char *name;
{
    register i = HashString(name, HASHSIZE);
    while (HashTable[i].Name)
	i = (i+1)%HASHSIZE;
    HashTable[i].Name = NewString(name);
    HashTable[i].Forgotten = 0;
}

static int Forget(name)
char *name;
{
    register i = HashString(name, HASHSIZE);
    while (HashTable[i].Name) {
	if (strcmp(HashTable[i].Name, name) == 0) {
	    HashTable[i].Forgotten = 1;
	    return 1;
	}
	i = (i+1)%HASHSIZE;
    }
    return 0;
}    

static int Lookup(name)
char *name;
{
    register i = HashString(name, HASHSIZE);
    while (HashTable[i].Name) {
	if (strcmp(HashTable[i].Name, name) == 0)
	    return 1;
	i = (i+1)%HASHSIZE;
    }
    return 0;
}
 
static char *FullPath(currentdir, name)
char *currentdir, *name;
{
    char path[MAXPATHLEN];
    if (*name == '/')
	return NewString(name);
    sprintf(path, "%s/%s", currentdir, name);
    return NewString(path);
}

static void makelinks(src, obj)
char *src;
char *obj;
{
    DIR *srcdir, *objdir;
    struct dirent *sp, *op;
    char desiredlinkvalue[MAXPATHLEN];
    register int i;

    objdir = opendir(obj);
    if (objdir == NULL)
	Stop("Cannot open object directory %s", obj);

    srcdir = opendir(src);
    if (srcdir == NULL || chdir(src) == -1)
	Stop("Cannot open source directory %s", src);

    while ((sp = readdir(srcdir)) != NULL) {
	struct stat status;
	if (stat(sp->d_name, &status) == -1) {
	    BTW("Cannot stat %s/%s, ignoring...", src, sp->d_name);
	    continue;
	}
	if ((status.st_mode & S_IFMT) == S_IFREG && *sp->d_name != '.') {
	    char *ext = (char *) rindex(sp->d_name, '.');
	    if (!ext || strcmp(ext, ".private"))
		Remember(sp->d_name);
	}
    }

    if (chdir(obj) == -1)
	Stop("Couldn't cd to %s", obj);
    while ((op = readdir(objdir)) != NULL) {
	struct stat status;
	int linkto, waslink, makelink;
	const char * removelink;

	removelink = 0; makelink = 0;
	linkto = Lookup(op->d_name);
	if (lstat(op->d_name, &status) == -1)
	    Stop("Couldn't stat %s/%s", obj, op->d_name);
	waslink = ((status.st_mode & S_IFMT) == S_IFLNK);
	sprintf(desiredlinkvalue, "%s/%s", src, op->d_name);

	if (!linkto && waslink) {
	    char *ext = (char *) rindex(op->d_name, '.');
	    removelink =
	      op->d_name[0] == '.'
		? "it's a \"dot\" file"
		: (ext && strcmp(ext, ".private") == 0)
		  ? "it was a link to a \".private\" file"
	    	  : "it referred to a nonexistent source file!";
	}

	if (linkto && !waslink) {
	    removelink = "it isn't a link, as it should be!";
	    makelink = 1;
	}

	if (linkto && waslink) {
	    char linkvalue[MAXPATHLEN];
	    int linksize = readlink(op->d_name, linkvalue, sizeof(linkvalue)-1);
	    if (linksize == -1)
		Stop("Unable to read contents of symbolic link %s/%s", obj, op->d_name);
	    linkvalue[linksize] = 0;
	    if (strcmp(desiredlinkvalue, linkvalue)) {
		removelink = "it's a link to the wrong place!";
		makelink = 1;
	    }
	}
	if (removelink) {
	    BTW("Removing %s/%s (%s)", obj, op->d_name, removelink);
	    if (unlink(op->d_name) == -1)
		Stop("Error removing %s/%s", obj, op->d_name);
	}

	if (makelink) {
	    BTW("and thus linking %s/%s to %s", obj, op->d_name, desiredlinkvalue);
	    if (symlink(desiredlinkvalue, op->d_name) == -1)
		Stop("Unable to create symbolic link %s/%s", obj, op->d_name);
	}
	Forget(op->d_name);
    }

    for (i = 0; i<HASHSIZE; i++) {
	if (HashTable[i].Name && ! HashTable[i].Forgotten) {
	    sprintf(desiredlinkvalue, "%s/%s", src, HashTable[i].Name);
	    BTW("Linking %s/%s to %s", obj, HashTable[i].Name, desiredlinkvalue);
	    if (symlink(desiredlinkvalue, HashTable[i].Name) == -1)
		Stop("Unable to create symbolic link %s/%s", obj, HashTable[i].Name);
	}
    }
    
    closedir(srcdir);
    closedir(objdir);
    
    ClearHashTable();

}

static void recursivemakelinks(src, obj)
char *src;
char *obj;
{
    struct stat stbuf;
    char newsrcdir[1000];
    char newobjdir[1000];
    char errorstr[1000];
    DIR *dirp;
    struct dirent *dirent;

    if (stat(obj, &stbuf) != 0) {
	if (errno == ENOENT) {
		if (mkdir(obj, 0777) != 0) {
			sprintf(errorstr, "rcsmakelinks: Couldn't create dir ``%s''", obj);
			perror(errorstr);
			return;
		}
	} else {
		sprintf(errorstr, "rcsmakelinks: Couldn't stat dir ``%s''", obj);
		perror(errorstr);
		return;
	}
    }
    BTW("Linking files in %s from %s\n", obj, src);
    makelinks(src,obj);
    if ((dirp = opendir(src)) == NULL) {
	sprintf(errorstr, "rcsmakelinks: Could not open dir %s", src);
	perror(errorstr);
	return;
    }
    while ((dirent = readdir(dirp)) != NULL) {
	if ((dirent->d_name[0] == '.')
	    ||(strcmp(dirent->d_name, "RCS") == 0)
	    ||(strcmp(dirent->d_name, "CVS.adm") == 0)) {
	    continue;	/* Ignore all hidden files and don't go down RCS links and don't try to link admin files */
	}
	sprintf(newsrcdir, "%s/%s", src, dirent->d_name);
	if (stat(newsrcdir, &stbuf)) {
	    sprintf(errorstr, "rcsmakelinks: Couldn't stat %s", newsrcdir);
	    perror(errorstr);
	    closedir(dirp);
	    return;
	}
	sprintf(newobjdir, "%s/%s", obj, dirent->d_name);
	if (stbuf.st_mode & S_IFDIR) {
	    recursivemakelinks(newsrcdir, newobjdir);
	}
    }
    closedir(dirp);
}

void main(argc,argv)
int argc;
char **argv;
{
    char *src, *obj;
    char cwd[MAXPATHLEN];
    int recurse = 0;

    if (argc >= 2 && (!strcmp(argv[1], "-r") || !strcmp(argv[1], "-R"))) {
	recurse = 1;
	--argc;
	++argv;
    }
    if (argc != 3)
	Stop("usage: rcsmakelinks [-R] srcdir objdir");
    src = argv[1]; obj = argv[2];

    if (*obj != '/' || *src != '/') {
	if (getwd(cwd) == -1)
	    Stop("Cannot determine working directory");
	obj = FullPath(cwd, obj);
	src = FullPath(cwd, src);
    }
    if (recurse) {
	recursivemakelinks(src, obj);
    } else {
	makelinks(src,obj);
    }
    exit(0);
}







